#version 130 

// Interpolated values from the vertex shaders
in vec3 PositionWorldPass;
in vec3 NormalWorldPass;
//in vec2 TexCoordPass;

// Ouput data
out vec4 Color;

// Values that stay constant for the whole mesh.
//uniform sampler2D gTextureSampler; // texture
uniform mat4 gModel, gView, gProjection;
uniform vec3 gEyePosition; // camera (in world)

struct Light
{
	vec4 Position;
	vec4 Ambient;
	vec4 Diffuse;
	vec4 Specular;

	float ConstantAttenuation;
	float LinearAttenuation;
	float QuadraticAttenuation;

	float SpotCutoff;
	float SpotExponent;
	vec3 SpotDirection;
};
uniform Light gLight;

struct Material
{
	vec4 Ambient;
	vec4 Diffuse;
	vec4 Specular;
	float Shininess;
};
uniform Material gMaterial;

void main()
{
	vec3 normalDirection = normalize(NormalWorldPass);
	vec3 viewDirection = normalize(gEyePosition - PositionWorldPass);

	vec3 lightDirection;
	float attenuation;
	float lightPower = 1.0f;

	if (gLight.Position.w == 0.0) // directional light
    {
		attenuation = 1.0; // no attenuation
		lightDirection = normalize(vec3(gLight.Position));
    }
	else // point or spot light (or other kind of light)
    {
		vec3 vertexToLightSource = gLight.Position.xyz - PositionWorldPass;
		float distance = length(vertexToLightSource);
		lightDirection = normalize(vertexToLightSource);
		attenuation = 1.0 / (gLight.ConstantAttenuation
                           + gLight.LinearAttenuation * distance
                           + gLight.QuadraticAttenuation * distance * distance);
 
		if (gLight.SpotCutoff <= 90.0) // spotlight
        {
			float clampedCosine = max(0.0, dot(-lightDirection, normalize(gLight.SpotDirection)));
			if (clampedCosine < cos(radians(gLight.SpotCutoff))) // outside of spotlight cone
            {
              attenuation = 0.0;
            }
			else
            {
              attenuation = attenuation * pow(clampedCosine, gLight.SpotExponent);
            }
        }
    }
 
	//vec4 materialDiffuse = texture2D(gTextureSampler, TexCoordPass) * gMaterial.Diffuse;
	vec4 materialDiffuse = gMaterial.Diffuse;

	vec3 ambientLighting = vec3(gLight.Ambient) * vec3(gMaterial.Ambient) * vec3(materialDiffuse);
 
	vec3 diffuseReflection = attenuation
		* lightPower
		* vec3(gLight.Diffuse) * vec3(materialDiffuse)
		* max(0.0, dot(normalDirection, lightDirection));
 
	vec3 specularReflection;
	if (dot(normalDirection, lightDirection) < 0.0) // light source on the wrong side?
    {
		specularReflection = vec3(0.0, 0.0, 0.0); // no specular reflection
    }
	else // light source on the right side
    {
		specularReflection = attenuation
			* lightPower
			* vec3(gLight.Specular) * vec3(gMaterial.Specular)
			* pow(max(0.0, dot(reflect(-lightDirection, normalDirection), viewDirection)),
              gMaterial.Shininess);
    }
 
	//Color = vec4(ambientLighting, 1.0);
	//Color = vec4(ambientLighting + diffuseReflection, 1.0);
	Color = vec4(ambientLighting + diffuseReflection + specularReflection, 1.0);
}