#include "moglclass.h"
#include "shaders.h"
#include "light.h"
#include "material.h"
#include "camera.h"

using namespace moglclass;

#define NUMSHADERS 4

Shader shShaders[NUMSHADERS];
Program spMain;

float g_fovy = 45.0f;
float g_aspect = 1.0f;
float g_zNear = 1.0f;
float g_zFar = 100.f;

Camera camera1(FLY);
//float cameraHeading = 0.0f;
//float cameraZ = 10.0f;
glm::vec3 g_eye(0, 0, 10);
glm::vec3 g_at(0, 0, 0);
glm::vec3 g_up(0, 1, 0);

glm::mat4 Identity(1.0f);
glm::mat4 World(1.0f);
glm::mat4 View(1.0f);
glm::mat4 Projection(1.0f);
glm::mat4 MVP(1.0f);
glm::mat4 xyRotation(1.0f);

#define MAX_COLOR_MODE 4
glm::vec4 teapotColor[MAX_COLOR_MODE] = { glm::vec4(1, 1, 1, 1), glm::vec4(0, 0, 1, 1), glm::vec4(1, 0, 1, 1), glm::vec4(1, 1, 0, 1) };
int g_colormode = 0;
bool g_wiremode = false;

Light light1(glm::vec4(4.0f, 4.0f, 4.0f, 0.0f), glm::vec4(0.1f, 0.1f, 0.1f, 1.0f), glm::vec4(1.0f, 1.0f, 1.0f, 1.0f), glm::vec4(1.0f, 1.0f, 1.0f, 1.0f), 1.0f, 0.0f, 0.0f, 180.0f, 0.0f, glm::vec3(0.0f, 0.0f, 0.0f));
Material teapotMaterial(glm::vec4(0.2f, 0.2f, 0.2f, 1.0f), glm::vec4(1.0f, 1.0f, 1.0f, 1.0f), glm::vec4(1.0f, 1.0f, 1.0f, 1.0f), 50.0f);
Material redMaterial(glm::vec4(0.2f, 0.2f, 0.2f, 1.0f), glm::vec4(1.0f, 0.0f, 0.0f, 1.0f), glm::vec4(1.0f, 1.0f, 1.0f, 1.0f), 50.0f);
Material greenMaterial(glm::vec4(0.2f, 0.2f, 0.2f, 1.0f), glm::vec4(0.0f, 1.0f, 0.0f, 1.0f), glm::vec4(1.0f, 1.0f, 1.0f, 1.0f), 50.0f);

void init( void )
{
    // load shaders and use the resulting shader program
	spMain.createProgram();
	shShaders[0].loadShader("GLUTPerPixelShading.vs", GL_VERTEX_SHADER);
	shShaders[1].loadShader("GLUTPerPixelShading.fs", GL_FRAGMENT_SHADER);
	spMain.attachShaderToProgram(&shShaders[0]);
	spMain.attachShaderToProgram(&shShaders[1]);
	spMain.linkProgram();
	spMain.useProgram();
	
	Projection = glm::perspective(g_fovy, g_aspect, g_zNear, g_zFar);
	View = camera1.lookAt(g_eye, g_at, g_up);

	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LESS);
	glClearColor( 1.0, 1.0, 1.0, 1.0 ); // white background
}

void cleanup()
{
	spMain.deleteProgram();
	printf("clean up\n");
}

void drawCylinder(glm::mat4& worldMatrix, glm::vec3& position)
{
	float radian = DEGTORAD(90.0f);
	//glm::mat4 R = glm::rotate(glm::mat4(1.0f), radian, glm::vec3(1, 0, 0));
	glm::mat4 R = glm::rotate(worldMatrix, radian, glm::vec3(1, 0, 0));
	World = glm::translate(R, position);
	spMain.setUniform("gModel", World);
	glutSolidCylinder(1.0, 5.0, 16, 4);
}

void drawSphere(glm::mat4& worldMatrix, glm::vec3& position)
{
	//World = glm::translate(glm::mat4(1.0f), position);
	World = glm::translate(worldMatrix, position);
	spMain.setUniform("gModel", World);
	glutSolidSphere(1.0, 16, 16);
}

void display( void )
{
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );     // clear the window

	// set drawmode
	if (g_wiremode) 
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);  // draw wireframe 
	else
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);	// draw filled 

	// MVP matrix
	View = camera1.View();
	//World = glm::rotate(glm::mat4(1.0f), -cameraHeading, glm::vec3(0, 1, 0)) * glm::translate(glm::mat4(1.0f), glm::vec3(0, 0, -cameraZ));

	MVP = Projection * View * World;
	spMain.useProgram();
	spMain.setUniform("gMVP", MVP);
	spMain.setUniform("gProjection", Projection);
	spMain.setUniform("gView", View);
	spMain.setUniform("gModel", World);

	glm::vec3 cameraPos = camera1.getPosition();
	//printf("cameraPos: %f %f %f\n", cameraPos[0], cameraPos[1], cameraPos[2]);
	spMain.setUniform("gEyePosition", cameraPos);

	// light
	light1.setUniform(&spMain, "gLight");

	// teapot material
	teapotMaterial.setDiffuse(teapotColor[g_colormode]);
	teapotMaterial.setUniform(&spMain, "gMaterial");
	// Teapot
	//World = glm::translate(glm::mat4(1.0f), glm::vec3(0, 0, 1));
	World = glm::translate(xyRotation, glm::vec3(0, 0, 1)); // teapot moving
	spMain.setUniform("gModel", World);
	glutSolidTeapot(1.5);

	// red material
	redMaterial.setUniform(&spMain, "gMaterial");
	// Cylinder
	drawCylinder(xyRotation, glm::vec3(3, 0, -1));
	drawCylinder(xyRotation, glm::vec3(-3, 0, -1));
	drawCylinder(xyRotation, glm::vec3(3, 3, -1));
	drawCylinder(xyRotation, glm::vec3(-3, 3, -1));
	drawCylinder(xyRotation, glm::vec3(3, -3, -1));
	drawCylinder(xyRotation, glm::vec3(-3, -3, -1));

	// green material
	greenMaterial.setUniform(&spMain, "gMaterial");
	// Sphere
	drawSphere(xyRotation, glm::vec3(-3, 3, 0));
	drawSphere(xyRotation, glm::vec3(3, 3, 0));
	drawSphere(xyRotation, glm::vec3(-3, 3, 3));
	drawSphere(xyRotation, glm::vec3(3, 3, 3));
	drawSphere(xyRotation, glm::vec3(-3, 3, -3));
	drawSphere(xyRotation, glm::vec3(3, 3, -3));

    glutSwapBuffers();
}

void update()
{
	static int currentTime, deltaTime, prevTime = 0; 
	static int startTime, elapsedTime = 0;
	currentTime = glutGet(GLUT_ELAPSED_TIME);
	deltaTime = currentTime - prevTime;
	prevTime = currentTime;

    // translate & rotate around x/y axis
	static float angle = 0.0f;
	angle += 0.001f;
	
	// geometry: translate & rotate in x & rotate in y
	glm::mat4 T = glm::translate(glm::mat4(1.0f), glm::vec3(0, 0, 1));
	glm::mat4 R = glm::rotate(T, angle, glm::vec3(1, 0, 0));
	xyRotation = glm::rotate(R, angle, glm::vec3(0, 1, 0));

	elapsedTime = currentTime - startTime;
	glutPostRedisplay();
}

void keyboard( unsigned char key, int x, int y )
{
	switch (key)  
	{    
	case 0x1B:    
	case 'q':    
	case 'Q':      
		exit(0);     
	case ' ':
		g_wiremode = !g_wiremode;
		break;  
	case 'c':
		g_colormode = (g_colormode+1) % 4;
		printf("g_colormode=%d\n", g_colormode);
		break;
	}
	glutPostRedisplay();
}
/*
void specialkey(int key, int x, int y )
{
	switch (key)  
	{    
	case GLUT_KEY_LEFT:
		cameraHeading -= 10;
		printf("cameraHeading=%f\n", cameraHeading);
		break;  
	case GLUT_KEY_RIGHT:    
		cameraHeading += 10;
		printf("cameraHeading=%f\n", cameraHeading);
		break;  
	case GLUT_KEY_UP:      
		cameraZ += 1.0f;
		printf("cameraZ=%f\n", cameraZ);
		break;  
	case GLUT_KEY_DOWN:
		cameraZ -= 1.0f;
		printf("cameraZ=%f\n", cameraZ);
		break;  
	case GLUT_KEY_PAGE_UP:      
		break;  
	case GLUT_KEY_PAGE_DOWN:
		break;  
	}
	glutPostRedisplay();
}
*/

void specialkey(int key, int x, int y)
{
 if (key == GLUT_KEY_F1)   // x-movement
        camera1.strafe(0.5);
 else if (key == GLUT_KEY_F2)
        camera1.strafe(-0.5);
 else if (key == GLUT_KEY_F3) // y-movement
        camera1.fly(0.5);
 else if (key == GLUT_KEY_F4)
        camera1.fly(-0.5);
 else if (key == GLUT_KEY_F5) // z-movement
        camera1.walk(0.5);
    else if (key == GLUT_KEY_F6)
        camera1.walk(-0.5);

 else if (key == GLUT_KEY_F7) // yaw (by y-axis)
        camera1.yaw(2.5);
 else if (key == GLUT_KEY_F8)
        camera1.yaw(-2.5);
 else if (key == GLUT_KEY_F9) // pitch (by x-axis)
        camera1.pitch(2.5);
    else if (key == GLUT_KEY_F10)
        camera1.pitch(-2.5);
 else if (key == GLUT_KEY_HOME) // roll (by z-axis)
        camera1.roll(2.5);
    else if (key == GLUT_KEY_END)
        camera1.roll(-2.5);

 else if (key == GLUT_KEY_LEFT) // same as town
        camera1.yaw(2.5);
 else if (key == GLUT_KEY_RIGHT)
        camera1.yaw(-2.5);
 else if (key == GLUT_KEY_UP) 
        camera1.walk(0.5);
    else if (key == GLUT_KEY_DOWN)
        camera1.walk(-0.5);

 glutPostRedisplay();
}


void mouse(int button, int state, int mx, int my)
{
	if (button==GLUT_RIGHT_BUTTON && state==GLUT_DOWN)
		exit(0);
}

void motion(int mx, int my)
{
	int w = glutGet(GLUT_WINDOW_WIDTH);
	int h = glutGet(GLUT_WINDOW_HEIGHT);
	float x = (float) (mx - w*0.5) / w;
	float y = (float) (h*0.5 - my) / h;
	//fprintf(stderr, "x=%f y=%f\n", x, y);

	glutPostRedisplay();
}

void reshape(int w, int h)
{
	g_aspect = (float) (w/h);

    glViewport(0, 0, (GLsizei) w, (GLsizei) h);
    glutPostRedisplay();
}

int main( int argc, char **argv )
{
	glutInit(&argc, argv);
    glutInitDisplayMode( GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH);
    glutInitWindowSize( SCREEN_WIDTH, SCREEN_HEIGHT );
    glutCreateWindow(argv[0]);
    
	glewInit();
    init();

    glutDisplayFunc( display );
    glutKeyboardFunc( keyboard );
    glutSpecialFunc( specialkey );
	glutMouseFunc( mouse ); 
	glutMotionFunc( motion ); 
	glutIdleFunc( update );
	glutReshapeFunc( reshape );

    glutMainLoop();

	cleanup();

    return 0;
}
